]> git.sesse.net Git - nageru/blobdiff - mixer.cpp
Add some GUI elements (hooked up) to help tuning the compressor.
[nageru] / mixer.cpp
index a91b3d8cc843c556dd376946eb05be9988e78e3a..861ea94d221bae78c93c06ec41f203aa30cd0905 100644 (file)
--- a/mixer.cpp
+++ b/mixer.cpp
@@ -124,7 +124,7 @@ Mixer::Mixer(const QSurfaceFormat &format, unsigned num_cards)
                        [this]{
                                resource_pool->clean_context();
                        });
-               card->resampler.reset(new Resampler(OUTPUT_FREQUENCY, OUTPUT_FREQUENCY, 2));
+               card->resampling_queue.reset(new ResamplingQueue(OUTPUT_FREQUENCY, OUTPUT_FREQUENCY, 2));
                card->usb->configure_card();
        }
 
@@ -153,6 +153,10 @@ Mixer::Mixer(const QSurfaceFormat &format, unsigned num_cards)
        r128.integr_start();
 
        locut.init(FILTER_HPF, 2);
+
+       // hlen=16 is pretty low quality, but we use quite a bit of CPU otherwise,
+       // and there's a limit to how important the peak meter is.
+       peak_resampler.setup(OUTPUT_FREQUENCY, OUTPUT_FREQUENCY * 4, /*num_channels=*/2, /*hlen=*/16);
 }
 
 Mixer::~Mixer()
@@ -182,10 +186,10 @@ int unwrap_timecode(uint16_t current_wrapped, int last)
        }
 }
 
-float find_peak(const vector<float> &samples)
+float find_peak(const float *samples, size_t num_samples)
 {
        float m = fabs(samples[0]);
-       for (size_t i = 1; i < samples.size(); ++i) {
+       for (size_t i = 1; i < num_samples; ++i) {
                m = std::max(m, fabs(samples[i]));
        }
        return m;
@@ -249,7 +253,7 @@ void Mixer::bm_frame(unsigned card_index, uint16_t timecode,
                if (dropped_frames > FPS * 2) {
                        fprintf(stderr, "Card %d lost more than two seconds (or time code jumping around), resetting resampler\n",
                                card_index);
-                       card->resampler.reset(new Resampler(OUTPUT_FREQUENCY, OUTPUT_FREQUENCY, 2));
+                       card->resampling_queue.reset(new ResamplingQueue(OUTPUT_FREQUENCY, OUTPUT_FREQUENCY, 2));
                } else if (dropped_frames > 0) {
                        // Insert silence as needed.
                        fprintf(stderr, "Card %d dropped %d frame(s) (before timecode 0x%04x), inserting silence.\n",
@@ -257,10 +261,10 @@ void Mixer::bm_frame(unsigned card_index, uint16_t timecode,
                        vector<float> silence;
                        silence.resize((OUTPUT_FREQUENCY / FPS) * 2);
                        for (int i = 0; i < dropped_frames; ++i) {
-                               card->resampler->add_input_samples((unwrapped_timecode - dropped_frames + i) / double(FPS), silence.data(), (OUTPUT_FREQUENCY / FPS));
+                               card->resampling_queue->add_input_samples((unwrapped_timecode - dropped_frames + i) / double(FPS), silence.data(), (OUTPUT_FREQUENCY / FPS));
                        }
                }
-               card->resampler->add_input_samples(unwrapped_timecode / double(FPS), audio.data(), num_samples);
+               card->resampling_queue->add_input_samples(unwrapped_timecode / double(FPS), audio.data(), num_samples);
        }
 
        // Done with the audio, so release it.
@@ -466,7 +470,7 @@ void Mixer::thread_func()
                for (unsigned card_index = 0; card_index < num_cards; ++card_index) {
                        input_frames.push_back(bmusb_current_rendering_frame[card_index]);
                }
-               const int64_t av_delay = TIMEBASE / 10;  // Corresponds to the fixed delay in resampler.h. TODO: Make less hard-coded.
+               const int64_t av_delay = TIMEBASE / 10;  // Corresponds to the fixed delay in resampling_queue.h. TODO: Make less hard-coded.
                h264_encoder->end_frame(fence, pts_int + av_delay, input_frames);
                ++frame;
                pts_int += TIMEBASE / FPS;
@@ -533,7 +537,7 @@ void Mixer::process_audio_one_frame()
                samples_card.resize((OUTPUT_FREQUENCY / FPS) * 2);
                {
                        unique_lock<mutex> lock(cards[card_index].audio_mutex);
-                       if (!cards[card_index].resampler->get_output_samples(pts(), &samples_card[0], OUTPUT_FREQUENCY / FPS)) {
+                       if (!cards[card_index].resampling_queue->get_output_samples(pts(), &samples_card[0], OUTPUT_FREQUENCY / FPS)) {
                                printf("Card %d reported previous underrun.\n", card_index);
                        }
                }
@@ -575,10 +579,12 @@ void Mixer::process_audio_one_frame()
 //     float limiter_att, compressor_att;
 
        // Then a limiter at +0 dB (so, -14 dBFS) to take out the worst peaks only.
-       {
-               float threshold = pow(10.0f, (ref_level_dbfs + 0.0f) / 20.0f);  // +0 dB.
-               float ratio = 1000.0f;  // Infinity.
-               float attack_time = 0.001f;
+       // Note that since ratio is not infinite, we could go slightly higher than this.
+       // Probably more tuning is warranted here.
+       if (limiter_enabled) {
+               float threshold = pow(10.0f, limiter_threshold_dbfs / 20.0f);
+               float ratio = 30.0f;
+               float attack_time = 0.0f;  // Instant.
                float release_time = 0.005f;
                float makeup_gain = 1.0f;  // 0 dB.
                limiter.process(samples_out.data(), samples_out.size() / 2, threshold, ratio, attack_time, release_time, makeup_gain);
@@ -586,20 +592,33 @@ void Mixer::process_audio_one_frame()
        }
 
        // Finally, the real compressor.
-       {
-               float threshold = pow(10.0f, (ref_level_dbfs - 12.0f) / 20.0f);  // -12 dB.
+       if (compressor_enabled) {
+               float threshold = pow(10.0f, compressor_threshold_dbfs / 20.0f);
                float ratio = 20.0f;
                float attack_time = 0.005f;
                float release_time = 0.040f;
-               float makeup_gain = 2.0f;  // +3 dB.
+               float makeup_gain = 2.0f;  // +6 dB.
                compressor.process(samples_out.data(), samples_out.size() / 2, threshold, ratio, attack_time, release_time, makeup_gain);
 //             compressor_att = compressor.get_attenuation();
        }
 
 //     printf("limiter=%+5.1f  compressor=%+5.1f\n", 20.0*log10(limiter_att), 20.0*log10(compressor_att));
 
-       // Find peak and R128 levels.
-       peak = max<float>(peak, find_peak(samples_out));
+       // Upsample 4x to find interpolated peak.
+       peak_resampler.inp_data = samples_out.data();
+       peak_resampler.inp_count = samples_out.size() / 2;
+
+       vector<float> interpolated_samples_out;
+       interpolated_samples_out.resize(samples_out.size());
+       while (peak_resampler.inp_count > 0) {  // About four iterations.
+               peak_resampler.out_data = &interpolated_samples_out[0];
+               peak_resampler.out_count = interpolated_samples_out.size() / 2;
+               peak_resampler.process();
+               size_t out_stereo_samples = interpolated_samples_out.size() / 2 - peak_resampler.out_count;
+               peak = max<float>(peak, find_peak(interpolated_samples_out.data(), out_stereo_samples * 2));
+       }
+
+       // Find R128 levels.
        vector<float> left, right;
        deinterleave_samples(samples_out, &left, &right);
        float *ptrs[] = { left.data(), right.data() };
@@ -696,6 +715,7 @@ void Mixer::channel_clicked(int preview_num)
 
 void Mixer::reset_meters()
 {
+       peak_resampler.reset();
        peak = 0.0f;
        r128.reset();
        r128.integr_start();